#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>


static const char GIF_SIGNATURE[] = "\x47\x49\x46"; /* "GIF" */
static const char GIF_VERSION[]   = "\x38\x39\x61"; /* "89a" */
static const int  GIF_HEADER_LENGTH = 13;
static const int  GIF_COLOR_TABLE_OFFSET = 10;
static const int  GIF_GLOBAL_COLOR_TABLE_BIT  = 0x80;
static const int  GIF_GLOBAL_COLOR_TABLE_BITS = 0x07;
static const int  GIF_RGB = 3;
static const char GIF_COMMENT_EXTENSION_BLOCK_ID[]   = "\x21\xfe";
static const int  GIF_BLOCK_TERMINATOR               = '\00';
static const int  GIF_COMMENT_EXTENSION_BLOCK_COMP_LEN = 2 + 1 + 1; /* + text */

void usage(void)
{
    printf("Usage: imgcmtgif TEXT < src.gif > dist.gif\n");
}

static int has_gif_signature(unsigned char *buff)
{
    return memcmp(GIF_SIGNATURE, buff, sizeof(GIF_SIGNATURE) - 1) == 0;
}

static int is_supported_version(unsigned char *buff)
{
    return memcmp(GIF_VERSION, buff + sizeof(GIF_SIGNATURE) - 1,
                  sizeof(GIF_VERSION) - 1) == 0;
}

static int get_global_color_table_length(unsigned char *header)
{
    unsigned int h = *(header + GIF_COLOR_TABLE_OFFSET);

    /* has global color table flag ? */
    if (!(h & GIF_GLOBAL_COLOR_TABLE_BIT))
        return 0;

    return 1 << ((h & GIF_GLOBAL_COLOR_TABLE_BITS) + 1);
}

static int add_gif_comment_block(unsigned char *buff, char *text)
{
    unsigned char *bp;
    unsigned int len;
   
    bp = buff;
    len = strlen(text);
    
    /* append GIF Comment extension block id(2) */
    strcpy(bp, GIF_COMMENT_EXTENSION_BLOCK_ID);
    bp += sizeof(GIF_COMMENT_EXTENSION_BLOCK_ID) - 1;
    /* append block length(1) */
    *bp++ = len;
    /* append block data(n) */
    strcpy(bp, text);
    bp += len;
    /* append? block terminator(1) */
    *bp++ = GIF_BLOCK_TERMINATOR;
 
    return bp - buff;
}

int main(int argc, char *argv[])
{
    unsigned char buff[8192];
    ssize_t rc;
    int color_table_length;
    char *text;
    unsigned long l;

    if (argc != 2) {
        usage();
        exit(1);
    }


    read(STDIN_FILENO, buff, GIF_HEADER_LENGTH);
    if (!has_gif_signature(buff)) {
        fprintf(stderr, "%s: is not GIF file\n", __FILE__);
        return 1;
    }
    if (!is_supported_version(buff)) {
        fprintf(stderr, "%s: is not GIF89a file\n", __FILE__);
        return 1;
    }
    /* skip GIF header */
    write(STDOUT_FILENO, buff, GIF_HEADER_LENGTH);
    /* skip Global color table */
    color_table_length = get_global_color_table_length(buff);
    if (color_table_length > 0) {
        read(STDIN_FILENO, buff, color_table_length * GIF_RGB);
        write(STDOUT_FILENO, buff, color_table_length * GIF_RGB);
    }

    /* create & inject Comment Extension block */
    l = strlen(argv[1]) + GIF_COMMENT_EXTENSION_BLOCK_COMP_LEN;
    text = malloc(l);
    add_gif_comment_block(text, argv[1]);
    write(STDOUT_FILENO, text, l);
    free(text);

    /* skip other block */
    while ((rc = read(STDIN_FILENO, buff, sizeof(buff))) > 0) {
        write(STDOUT_FILENO, buff, rc);
    }

    return 0;
}
