#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>

static const char JPEG_SOI_MARKER[]  = "\xff\xd8";   /* Start of Image */
static const char JPEG_COM_MARKER[]  = "\xff\xfe";   /* COM comment segment */
static const int  JPEG_SEGMENT_IDENTIFIES = 0xff;
static const int  JPEG_APPn_SEGMENT_TYPE  = 0xe0;    /* val & 0xf0 */
static const int  JPEG_STD_SEGMENTS_LEN = 6;
static const int  JPEG_SEGMENT_LEN_POS  = 2;
static const int  JPEG_MARKER_LEN       = 2;
static const int  JPEG_COM_SEGMENT_COMP_LEN = 2 + 2;  /* + body */

void usage(void)
{
    printf("Usage: imgcmtjpg TEXT < src.jpg > dist.jpg\n");
}

int has_jpeg_signature(unsigned char *buff)
{
    return memcmp(JPEG_SOI_MARKER, buff, sizeof(JPEG_SOI_MARKER) - 1) == 0;
}

int has_jpeg_appn_segment(unsigned char *buff)
{
    unsigned char *bp;
    int sid;

    bp = buff + sizeof(JPEG_SOI_MARKER) - 1;
    sid = *(bp + 1) & JPEG_APPn_SEGMENT_TYPE;

    if (*bp == JPEG_SEGMENT_IDENTIFIES && sid == JPEG_APPn_SEGMENT_TYPE)
        return 1;
    else
        return 0;
}

unsigned short get_jpeg_segment_length(unsigned char *segment)
{
    unsigned short l;

    l = *(segment + JPEG_SEGMENT_LEN_POS) << 8;
    l |= *(segment + JPEG_SEGMENT_LEN_POS + 1) & 0xff;

    if (l > 2)
        return l - 2;
    else
        return 0;
}

int add_jpeg_comment_segment(unsigned char *buff, char *text)
{
    unsigned char *bp;
    unsigned int len;

    bp = buff;
    len = strlen(text);

    /* append JPEG COM segment marker */
    strcpy(bp, JPEG_COM_MARKER);
    bp += sizeof(JPEG_COM_MARKER) - 1;
    /* append segment length(2) */
    *bp++ = (len + 2) >> 8;
    *bp++ = (len + 2);
    /* append segment data(n) */
    memcpy(bp, text, len);  /* without strings terminator('\0').*/

    return bp - buff;
}

int main(int argc, char *argv[])
{
    unsigned char buff[8192];
    ssize_t rc;
    char *text;
    unsigned short segment_l;
    size_t buf_l;
    unsigned long l;

    if (argc != 2) {
        usage();
        exit(1);
    }

    /* check JPEG standard segments */
    read(STDIN_FILENO, buff, JPEG_STD_SEGMENTS_LEN);
    if (!has_jpeg_signature(buff)) {
        fprintf(stderr, "%s: is not JPEG file\n", __FILE__);
        return 1;
    }

    if (!has_jpeg_appn_segment(buff)) {
        fprintf(stderr, "%s: is not JPEG/APPn segment\n", __FILE__);
    }

    /* skip JPEG SOI and APPn segments */
    write(STDOUT_FILENO, buff, JPEG_STD_SEGMENTS_LEN);
    /* skip APPn segment */
    segment_l = get_jpeg_segment_length(buff + JPEG_MARKER_LEN);
    buf_l = segment_l > sizeof(buff) ? sizeof(buff) : segment_l;
    while ((rc = read(STDIN_FILENO, buff, buf_l)) > 0) {
        write(STDOUT_FILENO, buff, rc);
        segment_l -= rc;
        buf_l = segment_l > sizeof(buff) ? sizeof(buff) : segment_l;
    }

    /* create & inject COM segment */
    l = strlen(argv[1]) + JPEG_COM_SEGMENT_COMP_LEN;
    text = malloc(l);
    add_jpeg_comment_segment(text, argv[1]);
    write(STDOUT_FILENO, text, l);
    free(text);

    /* skip other segment */
    while ((rc = read(STDIN_FILENO, buff, sizeof(buff))) > 0) {
        write(STDOUT_FILENO, buff, rc);
    }

    return 0;
}
